"""Represent a light."""
from pytradfri.color import supported_features
from pytradfri.const import ATTR_DEVICE_STATE, SUPPORT_BRIGHTNESS, \
    ATTR_LIGHT_DIMMER, SUPPORT_COLOR_TEMP, ATTR_LIGHT_MIREDS, \
    SUPPORT_HEX_COLOR, ATTR_LIGHT_COLOR_HEX, SUPPORT_XY_COLOR, \
    ATTR_LIGHT_COLOR_X, ATTR_LIGHT_COLOR_Y, ATTR_LIGHT_COLOR_HUE, \
    ATTR_LIGHT_COLOR_SATURATION, ATTR_LIGHT_CONTROL


class Light:
    """Represent a light.

    https://github.com/IPSO-Alliance/pub/blob/master/docs/IPSO-Smart-Objects.
    pdf
    """

    def __init__(self, device, index):
        self.device = device
        self.index = index

    @property
    def supported_features(self):
        return supported_features(self.raw)

    @property
    def state(self):
        return self.raw.get(ATTR_DEVICE_STATE) == 1

    @property
    def dimmer(self):
        if self.supported_features & SUPPORT_BRIGHTNESS:
            return self.raw.get(ATTR_LIGHT_DIMMER)

    @property
    def color_temp(self):
        if self.supported_features & SUPPORT_COLOR_TEMP:
            if self.raw.get(ATTR_LIGHT_MIREDS) != 0:
                return self.raw.get(ATTR_LIGHT_MIREDS)

    @property
    def hex_color(self):
        if self.supported_features & SUPPORT_HEX_COLOR:
            return self.raw.get(ATTR_LIGHT_COLOR_HEX)

    @property
    def xy_color(self):
        if self.supported_features & SUPPORT_XY_COLOR:
            return (self.raw.get(ATTR_LIGHT_COLOR_X),
                    self.raw.get(ATTR_LIGHT_COLOR_Y))

    @property
    def hsb_xy_color(self):
        return (self.raw.get(ATTR_LIGHT_COLOR_HUE),
                self.raw.get(ATTR_LIGHT_COLOR_SATURATION),
                self.raw.get(ATTR_LIGHT_DIMMER),
                self.raw.get(ATTR_LIGHT_COLOR_X),
                self.raw.get(ATTR_LIGHT_COLOR_Y))

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self.device.raw[ATTR_LIGHT_CONTROL][self.index]

    def __repr__(self):
        state = "on" if self.state else "off"
        return "<Light #{} - " \
               "name: {}, " \
               "state: {}, " \
               "dimmer: {}, "\
               "hex_color: {}, " \
               "xy_color: {}, " \
               "hsb_xy_color: {}, "\
               "supported features: {} " \
               ">".format(self.index, self.device.name, state, self.dimmer,
                          self.hex_color, self.xy_color,
                          self.hsb_xy_color, self.supported_features)



"""Class to control the blinds."""
from pytradfri.command import Command
from pytradfri.const import RANGE_BLIND, \
    ATTR_BLIND_CURRENT_POSITION, ATTR_START_BLINDS, ATTR_BLIND_TRIGGER
from pytradfri.device.blind import Blind
from pytradfri.device.controller import Controller


class BlindControl(Controller):
    """Class to control the blinds."""

    def __init__(self, device):
        self._device = device

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ATTR_START_BLINDS]

    @property
    def blinds(self):
        """Return blind objects of the blind control."""
        return [Blind(self._device, i) for i in range(len(self.raw))]

    def trigger_blind(self):
        """Trigger the blind's movement."""
        return self.set_value(
            {
                ATTR_BLIND_TRIGGER: True
            }
        )

    def set_state(self, state):
        """Set state of a blind."""
        self._value_validate(state, RANGE_BLIND, "Blind position")

        return self.set_value(
            {
                ATTR_BLIND_CURRENT_POSITION: state
            }
        )

    def set_value(self, value):
        """
        Set values on blind control.
        Returns a Command.
        """
        return Command('put', self._device.path, {
            ATTR_START_BLINDS:
                [
                    value
                ]
        })



"""Class to control the signal repeater."""
from pytradfri.const import ROOT_SIGNAL_REPEATER
from pytradfri.device.signal_repeater import SignalRepeater


class SignalRepeaterControl:
    """Class to control the signal repeaters."""

    def __init__(self, device):
        self._device = device

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ROOT_SIGNAL_REPEATER]

    @property
    def signal_repeaters(self):
        """Return signal repeater objects of the signal repeater control."""
        return [SignalRepeater(self._device, i) for i in range(len(self.raw))]

    def __repr__(self):
        return '<SignalRepeaterControl for {} ({} signal repeaters)>'.format(
            self._device.name,
            len(self.raw)
        )



"""Class to control the sockets."""
from pytradfri.command import Command
from pytradfri.const import ATTR_SWITCH_PLUG, ATTR_DEVICE_STATE
from pytradfri.device.socket import Socket


class SocketControl:
    """Class to control the sockets."""

    def __init__(self, device):
        self._device = device

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ATTR_SWITCH_PLUG]

    @property
    def sockets(self):
        """Return socket objects of the socket control."""
        return [Socket(self._device, i) for i in range(len(self.raw))]

    def set_state(self, state, *, index=0):
        """Set state of a socket."""
        return self.set_values({
            ATTR_DEVICE_STATE: int(state)
        }, index=index)

    def set_values(self, values, *, index=0):
        """
        Set values on socket control.
        Returns a Command.
        """
        assert len(self.raw) == 1, \
            'Only devices with 1 socket supported'

        return Command('put', self._device.path, {
            ATTR_SWITCH_PLUG: [
                values
            ]
        })

    def __repr__(self):
        return '<SocketControl for {} ({} sockets)>'.format(self._device.name,
                                                            len(self.raw))



"""Base class for a controller."""


class Controller:
    """Represent a controller."""

    def _value_validate(self, value, rnge, identifier="Given"):
        """
        Make sure a value is within a given range
        """
        if value is not None and (value < rnge[0] or value > rnge[1]):
            raise ValueError('%s value must be between %d and %d.'
                             % (identifier, rnge[0], rnge[1]))



"""Represent a blind."""
from pytradfri.const import ATTR_START_BLINDS, \
    ATTR_BLIND_CURRENT_POSITION


class Blind:
    """Represent a blind."""

    def __init__(self, device, index):
        self.device = device
        self.index = index

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self.device.raw[ATTR_START_BLINDS][self.index]

    @property
    def current_cover_position(self):
        """Get the current position of the blind."""
        return self.raw.get(ATTR_BLIND_CURRENT_POSITION)



"""Classes to interact with devices."""
from datetime import datetime

from pytradfri.const import ATTR_APPLICATION_TYPE, ROOT_DEVICES, \
    ATTR_LAST_SEEN, ATTR_REACHABLE_STATE, \
    ATTR_LIGHT_CONTROL, ATTR_SWITCH_PLUG, ATTR_DEVICE_INFO,\
    ATTR_START_BLINDS, ROOT_SIGNAL_REPEATER
from pytradfri.device.blind_control import BlindControl
from pytradfri.device.light_control import LightControl
from pytradfri.device.signal_repeater_control import SignalRepeaterControl
from pytradfri.device.socket_control import SocketControl
from pytradfri.resource import ApiResource


class Device(ApiResource):
    """Base class for devices."""

    @property
    def application_type(self):
        return self.raw.get(ATTR_APPLICATION_TYPE)

    @property
    def path(self):
        return [ROOT_DEVICES, self.id]

    @property
    def device_info(self):
        return DeviceInfo(self)

    @property
    def last_seen(self):
        if ATTR_LAST_SEEN not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_LAST_SEEN])

    @property
    def reachable(self):
        return self.raw.get(ATTR_REACHABLE_STATE) == 1

    @property
    def has_light_control(self):
        return (self.raw is not None and
                len(self.raw.get(ATTR_LIGHT_CONTROL, "")) > 0)

    @property
    def light_control(self):
        return LightControl(self)

    @property
    def has_socket_control(self):
        return (self.raw is not None and
                len(self.raw.get(ATTR_SWITCH_PLUG, "")) > 0)

    @property
    def socket_control(self):
        if self.has_socket_control:
            return SocketControl(self)

    @property
    def has_blind_control(self):
        return (self.raw is not None and
                len(self.raw.get(ATTR_START_BLINDS, "")) > 0)

    @property
    def blind_control(self):
        if self.has_blind_control:
            return BlindControl(self)

    @property
    def has_signal_repeater_control(self):
        return (self.raw is not None and
                len(self.raw.get(ROOT_SIGNAL_REPEATER, "")) > 0)

    @property
    def signal_repeater_control(self):
        if self.has_signal_repeater_control:
            return SignalRepeaterControl(self)

    def __repr__(self):
        return "<{} - {} ({})>".format(self.id, self.name,
                                       self.device_info.model_number)


class DeviceInfo:
    """Represent device information.

    http://www.openmobilealliance.org/tech/profiles/LWM2M_Device-v1_0.xml
    """
    ATTR_MANUFACTURER = "0"
    ATTR_MODEL_NUMBER = "1"
    ATTR_SERIAL = "2"
    ATTR_FIRMWARE_VERSION = "3"
    ATTR_POWER_SOURCE = "6"
    VALUE_POWER_SOURCES = {
        1: 'Internal Battery',
        2: 'External Battery',
        3: 'Battery',  # Not in spec, used by remote
        4: 'Power over Ethernet',
        5: 'USB',
        6: 'AC (Mains) power',
        7: 'Solar'
    }
    ATTR_BATTERY = "9"

    def __init__(self, device):
        self._device = device

    @property
    def manufacturer(self):
        """Human readable manufacturer name."""
        return self.raw.get(DeviceInfo.ATTR_MANUFACTURER)

    @property
    def model_number(self):
        """A model identifier string (manufactuer specified string)."""
        return self.raw.get(DeviceInfo.ATTR_MODEL_NUMBER)

    @property
    def serial(self):
        """Serial number string."""
        return self.raw.get(DeviceInfo.ATTR_SERIAL)

    @property
    def firmware_version(self):
        """Current firmware version string of the device."""
        return self.raw.get(DeviceInfo.ATTR_FIRMWARE_VERSION)

    @property
    def power_source(self):
        """Power source."""
        return self.raw.get(DeviceInfo.ATTR_POWER_SOURCE)

    @property
    def power_source_str(self):
        """String representation of current power source."""
        if DeviceInfo.ATTR_POWER_SOURCE not in self.raw:
            return None
        return DeviceInfo.VALUE_POWER_SOURCES.get(self.power_source, 'Unknown')

    @property
    def battery_level(self):
        """Battery in 0..100"""
        return self.raw.get(DeviceInfo.ATTR_BATTERY)

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ATTR_DEVICE_INFO]
